# -*- coding: utf-8 -*-
"""
    { Settings }
"""
import pathlib

# Base Directory
BASE_DIR = pathlib.Path(__file__).parents[1]

# Installed Apps
# INSTALLED_APPS = ["good_app", "demo"]
INSTALLED_APPS = ["shop"]

SQL_URL = "sqlite:///example.db"
MONGO_URL = "mongodb://localhost:27017/test_database"



"""
    Fastberry Main (Core-Methods)
"""
from starlette.middleware.base import BaseHTTPMiddleware as BaseMiddleware
from strawberry.extensions import Extension as BaseExtension
from strawberry.permission import BasePermission

import spoc

# FrameWork
from .components import APIRouter as Router
from .components import cli
from .components import graphql as gql
from .framework import Fastberry as App

# GraphQL Tools
from .graphql import edges, error, errors, mutation, page, query

# GraphQL Premade User-Inputs
from .tools import Item as item
from .tools import Pagination as pagination


# Framework Wrappers
base_dir = spoc.base_dir
config = spoc.config
mode = spoc.mode
project = spoc.project
settings = spoc.settings

# Tools
component = spoc.component

try:
    import dbcontroller as dbc
    from dbcontroller.forms import ISNULL

    if hasattr(settings, "SQL_URL"):
        if settings.SQL_URL is not None:
            sql = dbc.Controller(sql=settings.SQL_URL)
    if hasattr(settings, "MONGO_URL"):
        if settings.MONGO_URL is not None:
            mongo = dbc.Controller(mongo=settings.MONGO_URL)

    # Types
    type = dbc.type

    # Forms
    input = dbc.form.graphql
    value = dbc.form.field

    # Value Tool
    filters = dbc.form.filters

    # Types Tool (DBController)
    field = dbc.field
    manager = dbc.manager

    # Scalars
    ID = dbc.ID
    date = dbc.date
    datetime = dbc.datetime
    time = dbc.time
    decimal = dbc.decimal
    text = dbc.text
    time = dbc.time
    json = dbc.json

    # Tester
    Date = dbc.Date

except ImportError:
    import strawberry

    # Types
    type = strawberry.type
    input = strawberry.input



# -*- coding: utf-8 -*-
""" [Main]
    FastAPI's Main File.
"""

# FastAPI & Strawberry
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from fastberry.app import Controller
from strawberry.fastapi import GraphQLRouter

"""
__________               __________                   __               __   
\______   \ ___________  \______   \_______  ____    |__| ____   _____/  |_ 
 |     ___// __ \_  __ \  |     ___/\_  __ \/  _ \   |  |/ __ \_/ ___\   __\
 |    |   \  ___/|  | \/  |    |     |  | \(  <_> )  |  \  ___/\  \___|  |  
 |____|    \___  >__|     |____|     |__|   \____/\__|  |\___  >\___  >__|  
               \/                                \______|    \/     \/      
"""

"""
import fastberry as fb
from sqlalchemy import create_engine

# print(Controller.types)
engine = create_engine(fb.sql.url, echo=True)
fb.sql.base.metadata.create_all(engine)
"""


# FastAPI
app = Controller.context.app

# Global Variables
SPOC_ALLOWED_HOSTS = Controller.toml.get("api", {}).get("allowed_hosts")
SPOC_MAX_DEPTH = Controller.toml.get("api", {}).get("max_depth", 3)
SPOC_GRAPHQL_PATH = Controller.toml.get("api", {}).get("graphql_path", "graphql")

# (SPOC_GRAPHQL_PATH) Global Variables
if SPOC_GRAPHQL_PATH.startswith("/"):
    SPOC_GRAPHQL_PATH = SPOC_GRAPHQL_PATH[1:]

# Middleware
app.add_middleware(GZipMiddleware, minimum_size=1000)
if Controller.extras and Controller.extras.middleware:
    Controller.extras.middleware(app)

# CORS
if SPOC_ALLOWED_HOSTS:
    app.add_middleware(
        CORSMiddleware,
        allow_origins=SPOC_ALLOWED_HOSTS,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

# ALL Routers
if Controller.router:
    app.include_router(*Controller.router)

# GraphQL (Config)
if Controller.graphql:
    schema = Controller.graphql.schema(
        introspection=(not Controller.context.is_production),
        max_depth=SPOC_MAX_DEPTH,
        extensions=Controller.extras.extensions,
    )

    # GraphQL (Schema)
    if schema:
        app.include_router(
            GraphQLRouter(
                schema,
                graphiql=(not Controller.context.is_production),
            ),
            prefix=f"/{SPOC_GRAPHQL_PATH}",
            tags=["GraphQL"],
        )

# Startup
@app.on_event("startup")
async def startup_event():
    """Startup Methods"""
    if Controller.extras and Controller.extras.on_startup:
        for func in Controller.extras.on_startup:
            if func:
                func(Controller.context)


# Shutdown
@app.on_event("shutdown")
def shutdown_event():
    """Shutdown Methods"""
    if Controller.extras and Controller.extras.on_shutdown:
        for func in Controller.extras.on_shutdown:
            if func:
                func(Controller.context)



"""
    Types
"""

try:
    import dbcontroller as dbc
except ImportError:
    dbc = None


def types(models: list):
    """Collect (GraphQL) Return-Types"""
    single_name = {}
    all_types = {}
    if dbc:
        for active in models:
            is_component = dbc.is_model(active.object)
            if is_component:
                if hasattr(active.object, "__database__"):
                    name = active.object.__name__
                    found = single_name.get(name)
                    if found and found.object != active.object:
                        found_message = f"""\n* <Type: { found.name }> is already in use by the <App: { found.app }>"""
                        found_message += f"""\n* <Types> must have UNIQUE names."""
                        found_message += f"""\n* Rename it in either <App: { found.app } or { active.app }>."""
                        raise ValueError(found_message)
                    single_name[name] = active
                    if callable(active.object.__database__):
                        active.object.__database__()
                    all_types[active.key] = active.object
    return all_types
